@@ -15,7 +15,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype Shader
-    \inherits Object
+    \inherits QtObject
     \inqmlmodule QtQuick3D
     \brief Container component for defining shader code used by post-processing effects.
 
@@ -51,7 +51,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype TextureInput
-    \inherits Object
+    \inherits QtObject
     \inqmlmodule QtQuick3D
     \brief Specifies a texture exposed to the shaders of a CustomMaterial or Effect.
 
@@ -77,7 +77,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype Pass
-    \inherits Object
+    \inherits QtObject
     \inqmlmodule QtQuick3D
     \brief Defines a render pass in an Effect.
 
@@ -104,7 +104,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype Command
-    \inherits Object
+    \inherits QtObject
     \inqmlmodule QtQuick3D
     \brief Supertype of commands to be performed as part of a pass in an Effect.
 
@@ -143,7 +143,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype Buffer
-    \inherits Object
+    \inherits QtObject
     \inqmlmodule QtQuick3D
     \brief Creates or references a color buffer to be used for a pass of an Effect.
 
@@ -289,105 +289,105 @@ static constexpr QMetaType::Type qssg_metatype_list[] {
 template<>
 struct ShaderType<QMetaType::Double>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Float; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Float; }
     static QByteArray name() { return QByteArrayLiteral("float"); }
 };
 
 template<>
 struct ShaderType<QMetaType::Bool>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Boolean; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Boolean; }
     static QByteArray name() { return QByteArrayLiteral("bool"); }
 };
 
 template<>
 struct ShaderType<QMetaType::Int>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Integer; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Integer; }
     static QByteArray name() { return QByteArrayLiteral("int"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QVector2D>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Vec2; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Vec2; }
     static QByteArray name() { return QByteArrayLiteral("vec2"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QVector3D>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Vec3; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Vec3; }
     static QByteArray name() { return QByteArrayLiteral("vec3"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QVector4D>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Vec4; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Vec4; }
     static QByteArray name() { return QByteArrayLiteral("vec4"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QColor>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Rgba; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Rgba; }
     static QByteArray name() { return QByteArrayLiteral("vec4"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QSize>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Size; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Size; }
     static QByteArray name() { return QByteArrayLiteral("vec2"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QSizeF>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::SizeF; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::SizeF; }
     static QByteArray name() { return QByteArrayLiteral("vec2"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QPoint>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Point; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Point; }
     static QByteArray name() { return QByteArrayLiteral("vec2"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QPointF>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::PointF; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::PointF; }
     static QByteArray name() { return QByteArrayLiteral("vec2"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QRect>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Rect; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Rect; }
     static QByteArray name() { return QByteArrayLiteral("vec4"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QRectF>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::RectF; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::RectF; }
     static QByteArray name() { return QByteArrayLiteral("vec4"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QQuaternion>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Quaternion; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Quaternion; }
     static QByteArray name() { return QByteArrayLiteral("vec4"); }
 };
 
 template<>
 struct ShaderType<QMetaType::QMatrix4x4>
 {
-    static constexpr QSSGRenderShaderDataType type() { return QSSGRenderShaderDataType::Matrix4x4; }
+    static constexpr QSSGRenderShaderValue::Type type() { return QSSGRenderShaderValue::Matrix4x4; }
     static QByteArray name() { return QByteArrayLiteral("mat4"); }
 };
 
@@ -430,45 +430,45 @@ QByteArray uniformTypeName(QMetaType type)
     }
 }
 
-QByteArray uniformTypeName(QSSGRenderShaderDataType type)
+QByteArray uniformTypeName(QSSGRenderShaderValue::Type type)
 {
     switch (type) {
-    case QSSGRenderShaderDataType::Float:
+    case QSSGRenderShaderValue::Float:
         return ShaderType<QMetaType::Double>::name();
-    case QSSGRenderShaderDataType::Boolean:
+    case QSSGRenderShaderValue::Boolean:
         return ShaderType<QMetaType::Bool>::name();
-    case QSSGRenderShaderDataType::Integer:
+    case QSSGRenderShaderValue::Integer:
         return ShaderType<QMetaType::Int>::name();
-    case QSSGRenderShaderDataType::Vec2:
+    case QSSGRenderShaderValue::Vec2:
         return ShaderType<QMetaType::QVector2D>::name();
-    case QSSGRenderShaderDataType::Vec3:
+    case QSSGRenderShaderValue::Vec3:
         return ShaderType<QMetaType::QVector3D>::name();
-    case QSSGRenderShaderDataType::Vec4:
+    case QSSGRenderShaderValue::Vec4:
         return ShaderType<QMetaType::QVector4D>::name();
-    case QSSGRenderShaderDataType::Rgba:
+    case QSSGRenderShaderValue::Rgba:
         return ShaderType<QMetaType::QColor>::name();
-    case QSSGRenderShaderDataType::Size:
+    case QSSGRenderShaderValue::Size:
         return ShaderType<QMetaType::QSize>::name();
-    case QSSGRenderShaderDataType::SizeF:
+    case QSSGRenderShaderValue::SizeF:
         return ShaderType<QMetaType::QSizeF>::name();
-    case QSSGRenderShaderDataType::Point:
+    case QSSGRenderShaderValue::Point:
         return ShaderType<QMetaType::QPoint>::name();
-    case QSSGRenderShaderDataType::PointF:
+    case QSSGRenderShaderValue::PointF:
         return ShaderType<QMetaType::QPointF>::name();
-    case QSSGRenderShaderDataType::Rect:
+    case QSSGRenderShaderValue::Rect:
         return ShaderType<QMetaType::QRect>::name();
-    case QSSGRenderShaderDataType::RectF:
+    case QSSGRenderShaderValue::RectF:
         return ShaderType<QMetaType::QRectF>::name();
-    case QSSGRenderShaderDataType::Quaternion:
+    case QSSGRenderShaderValue::Quaternion:
         return ShaderType<QMetaType::QQuaternion>::name();
-    case QSSGRenderShaderDataType::Matrix4x4:
+    case QSSGRenderShaderValue::Matrix4x4:
         return ShaderType<QMetaType::QMatrix4x4>::name();
     default:
         return QByteArray();
     }
 }
 
-QSSGRenderShaderDataType uniformType(QMetaType type)
+QSSGRenderShaderValue::Type uniformType(QMetaType type)
 {
     switch (type.id()) {
     case QMetaType::Double:
@@ -503,7 +503,7 @@ QSSGRenderShaderDataType uniformType(QMetaType type)
     case QMetaType::QMatrix4x4:
         return ShaderType<QMetaType::QMatrix4x4>::type();
     default:
-        return QSSGRenderShaderDataType::Unknown;
+        return QSSGRenderShaderValue::Unknown;
     }
 }
 
@@ -613,6 +613,11 @@ void QQuick3DShaderUtilsRenderPass::qmlAppendShader(QQmlListProperty<QQuick3DSha
 
     // the only thing we can do is to append to our list, do not try to be clever
     that->m_shaders.append(shader);
+
+    connect(shader, &QQuick3DShaderUtilsShader::shaderChanged, that, &QQuick3DShaderUtilsRenderPass::changed);
+    connect(shader, &QQuick3DShaderUtilsShader::stageChanged, that, &QQuick3DShaderUtilsRenderPass::changed);
+
+    emit that->changed();
 }
 
 QQuick3DShaderUtilsShader *QQuick3DShaderUtilsRenderPass::qmlShaderAt(QQmlListProperty<QQuick3DShaderUtilsShader> *list,
@@ -631,7 +636,13 @@ qsizetype QQuick3DShaderUtilsRenderPass::qmlShaderCount(QQmlListProperty<QQuick3
 void QQuick3DShaderUtilsRenderPass::qmlShaderClear(QQmlListProperty<QQuick3DShaderUtilsShader> *list)
 {
     QQuick3DShaderUtilsRenderPass *that = qobject_cast<QQuick3DShaderUtilsRenderPass *>(list->object);
+
+    for (QQuick3DShaderUtilsShader *shader : that->m_shaders)
+        shader->disconnect(that);
+
     that->m_shaders.clear();
+
+    emit that->changed();
 }
 
 QQmlListProperty<QQuick3DShaderUtilsShader> QQuick3DShaderUtilsRenderPass::shaders()
@@ -659,22 +670,18 @@ void QQuick3DShaderUtilsTextureInput::setTexture(QQuick3DTexture *texture)
     while (p != nullptr) {
         if (QQuick3DCustomMaterial *mat = qobject_cast<QQuick3DCustomMaterial *>(p)) {
             mat->setDynamicTextureMap(this);
-            QQuick3DObjectPrivate::updatePropertyListener(texture, m_texture, QQuick3DObjectPrivate::get(mat)->sceneManager, name, mat->m_connections, [this](QQuick3DObject *) {
-                setTexture(nullptr);
-            });
+            QQuick3DObjectPrivate::attachWatcherPriv(mat, this, &QQuick3DShaderUtilsTextureInput::setTexture, texture, m_texture);
             break;
         } else if (QQuick3DEffect *efx = qobject_cast<QQuick3DEffect *>(p)) {
             efx->setDynamicTextureMap(this);
-            QQuick3DObjectPrivate::updatePropertyListener(texture, m_texture, QQuick3DObjectPrivate::get(efx)->sceneManager, name, efx->m_connections, [this](QQuick3DObject *) {
-                setTexture(nullptr);
-            });
+            QQuick3DObjectPrivate::attachWatcherPriv(efx, this, &QQuick3DShaderUtilsTextureInput::setTexture, texture, m_texture);
             break;
         }
         p = p->parent();
     }
 
     if (p == nullptr) {
-        qWarning("A texture was defined out of Material or Effect");
+        qWarning("A TextureInput was defined without a CustomMaterial or Effect ancestor. This should be avoided.");
     }
 
     m_texture = texture;
