@@ -10,6 +10,7 @@
 
 #include <QtQuick3DRuntimeRender/private/qssgrenderer_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendercamera_p.h>
+#include <QtQuick3DRuntimeRender/private/qssglayerrenderdata_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -33,44 +34,55 @@ struct ParticleLightData
     float spotLightInnerConeAngle[4] = {0.0f};
 };
 
-void QSSGParticleRenderer::updateUniformsForParticles(QSSGRef<QSSGRhiShaderPipeline> &shaders,
+void QSSGParticleRenderer::updateUniformsForParticles(QSSGRhiShaderPipeline &shaders,
                                                       QSSGRhiContext *rhiCtx,
                                                       char *ubufData,
                                                       QSSGParticlesRenderable &renderable,
-                                                      QSSGRenderCamera &inCamera)
+                                                      const QSSGRenderCameraList &cameras)
 {
     const QMatrix4x4 clipSpaceCorrMatrix = rhiCtx->rhi()->clipSpaceCorrMatrix();
 
-    QSSGRhiShaderPipeline::CommonUniformIndices &cui = shaders->commonUniformIndices;
+    QSSGRhiShaderPipeline::CommonUniformIndices &cui = shaders.commonUniformIndices;
 
-    const QMatrix4x4 projection = clipSpaceCorrMatrix * inCamera.projection;
-    shaders->setUniform(ubufData, "qt_projectionMatrix", projection.constData(), 16 * sizeof(float), &cui.projectionMatrixIdx);
-
-    const QMatrix4x4 viewMatrix = inCamera.globalTransform.inverted();
-    shaders->setUniform(ubufData, "qt_viewMatrix", viewMatrix.constData(), 16 * sizeof(float), &cui.viewMatrixIdx);
+    const int viewCount = cameras.count();
+    if (viewCount < 2) {
+        const QMatrix4x4 projection = clipSpaceCorrMatrix * cameras[0]->projection;
+        shaders.setUniform(ubufData, "qt_projectionMatrix", projection.constData(), 16 * sizeof(float), &cui.projectionMatrixIdx);
+        const QMatrix4x4 viewMatrix = cameras[0]->globalTransform.inverted();
+        shaders.setUniform(ubufData, "qt_viewMatrix", viewMatrix.constData(), 16 * sizeof(float), &cui.viewMatrixIdx);
+    } else {
+        QVarLengthArray<QMatrix4x4, 2> projectionMatrices(viewCount);
+        QVarLengthArray<QMatrix4x4, 2> viewMatrices(viewCount);
+        for (int viewIndex = 0; viewIndex < viewCount; ++viewIndex) {
+            projectionMatrices[viewIndex] = clipSpaceCorrMatrix * cameras[viewIndex]->projection;
+            viewMatrices[viewIndex] = cameras[viewIndex]->globalTransform.inverted();
+        }
+        shaders.setUniformArray(ubufData, "qt_projectionMatrix", projectionMatrices.constData(), viewCount, QSSGRenderShaderValue::Matrix4x4, &cui.projectionMatrixIdx);
+        shaders.setUniformArray(ubufData, "qt_viewMatrix", viewMatrices.constData(), viewCount, QSSGRenderShaderValue::Matrix4x4, &cui.viewMatrixIdx);
+    }
 
     const QMatrix4x4 &modelMatrix = renderable.globalTransform;
-    shaders->setUniform(ubufData, "qt_modelMatrix", modelMatrix.constData(), 16 * sizeof(float), &cui.modelMatrixIdx);
+    shaders.setUniform(ubufData, "qt_modelMatrix", modelMatrix.constData(), 16 * sizeof(float), &cui.modelMatrixIdx);
 
     QVector2D oneOverSize = QVector2D(1.0f, 1.0f);
     auto &particleBuffer = renderable.particles.m_particleBuffer;
     const quint32 particlesPerSlice = particleBuffer.particlesPerSlice();
     oneOverSize = QVector2D(1.0f / particleBuffer.size().width(), 1.0f / particleBuffer.size().height());
-    shaders->setUniform(ubufData, "qt_oneOverParticleImageSize", &oneOverSize, 2 * sizeof(float));
-    shaders->setUniform(ubufData, "qt_countPerSlice", &particlesPerSlice, 1 * sizeof(quint32));
+    shaders.setUniform(ubufData, "qt_oneOverParticleImageSize", &oneOverSize, 2 * sizeof(float));
+    shaders.setUniform(ubufData, "qt_countPerSlice", &particlesPerSlice, 1 * sizeof(quint32));
 
     // Global opacity of the particles node
-    shaders->setUniform(ubufData, "qt_opacity", &renderable.opacity, 1 * sizeof(float));
+    shaders.setUniform(ubufData, "qt_opacity", &renderable.opacity, 1 * sizeof(float));
 
     float blendImages = renderable.particles.m_blendImages ? 1.0f : 0.0f;
     float imageCount = float(renderable.particles.m_spriteImageCount);
     float ooImageCount = 1.0f / imageCount;
 
     QVector4D spriteConfig(imageCount, ooImageCount, 0.0f, blendImages);
-    shaders->setUniform(ubufData, "qt_spriteConfig", &spriteConfig, 4 * sizeof(float));
+    shaders.setUniform(ubufData, "qt_spriteConfig", &spriteConfig, 4 * sizeof(float));
 
     const float billboard = renderable.particles.m_billboard ? 1.0f : 0.0f;
-    shaders->setUniform(ubufData, "qt_billboard", &billboard, 1 * sizeof(float));
+    shaders.setUniform(ubufData, "qt_billboard", &billboard, 1 * sizeof(float));
 
     // Lights
     QVector3D theLightAmbientTotal;
@@ -86,51 +98,46 @@ void QSSGParticleRenderer::updateUniformsForParticles(QSSGRef<QSSGRhiShaderPipel
             // Ignore lights which are not specified for the particle
             if (!renderable.particles.m_lights.contains(theLight))
                 continue;
-            const bool lightEnabled = lights[lightIdx].enabled;
-            if (lightEnabled) {
-                if (theLight->m_brightness > 0.0f) {
-                    if (theLight->type == QSSGRenderLight::Type::DirectionalLight) {
-                        theLightAmbientTotal += theLight->m_diffuseColor * theLight->m_brightness;
-                    } else if (theLight->type == QSSGRenderLight::Type::PointLight && pointLight < 4) {
-                        lightData.pointLightColor[pointLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
-                        lightData.pointLightPos[pointLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
-                        lightData.pointLightConstantAtt[pointLight] = aux::translateConstantAttenuation(theLight->m_constantFade);
-                        lightData.pointLightLinearAtt[pointLight] = aux::translateLinearAttenuation(theLight->m_linearFade);
-                        lightData.pointLightQuadAtt[pointLight] = aux::translateQuadraticAttenuation(theLight->m_quadraticFade);
-                        pointLight++;
-                    } else if (theLight->type == QSSGRenderLight::Type::SpotLight && spotLight < 4) {
-                        lightData.spotLightColor[spotLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
-                        lightData.spotLightPos[spotLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
-                        lightData.spotLightDir[spotLight] = QVector4D(lights[lightIdx].direction, 0.0f);
-                        lightData.spotLightConstantAtt[spotLight] = aux::translateConstantAttenuation(theLight->m_constantFade);
-                        lightData.spotLightLinearAtt[spotLight] = aux::translateLinearAttenuation(theLight->m_linearFade);
-                        lightData.spotLightQuadAtt[spotLight] = aux::translateQuadraticAttenuation(theLight->m_quadraticFade);
-                        float coneAngle = theLight->m_coneAngle;
-                        // Inner cone angle must always be < cone angle, to not have possible undefined behavior for shader smoothstep
-                        float innerConeAngle = std::min(theLight->m_innerConeAngle, coneAngle - 0.01f);
-                        lightData.spotLightConeAngle[spotLight] = qDegreesToRadians(coneAngle);
-                        lightData.spotLightInnerConeAngle[spotLight] = qDegreesToRadians(innerConeAngle);
-                        spotLight++;
-                    }
-                }
-                theLightAmbientTotal += theLight->m_ambientColor;
+            if (theLight->type == QSSGRenderLight::Type::DirectionalLight) {
+                theLightAmbientTotal += theLight->m_diffuseColor * theLight->m_brightness;
+            } else if (theLight->type == QSSGRenderLight::Type::PointLight && pointLight < 4) {
+                lightData.pointLightColor[pointLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
+                lightData.pointLightPos[pointLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
+                lightData.pointLightConstantAtt[pointLight] = QSSGUtils::aux::translateConstantAttenuation(theLight->m_constantFade);
+                lightData.pointLightLinearAtt[pointLight] = QSSGUtils::aux::translateLinearAttenuation(theLight->m_linearFade);
+                lightData.pointLightQuadAtt[pointLight] = QSSGUtils::aux::translateQuadraticAttenuation(theLight->m_quadraticFade);
+                pointLight++;
+            } else if (theLight->type == QSSGRenderLight::Type::SpotLight && spotLight < 4) {
+                lightData.spotLightColor[spotLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
+                lightData.spotLightPos[spotLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
+                lightData.spotLightDir[spotLight] = QVector4D(lights[lightIdx].direction, 0.0f);
+                lightData.spotLightConstantAtt[spotLight] = QSSGUtils::aux::translateConstantAttenuation(theLight->m_constantFade);
+                lightData.spotLightLinearAtt[spotLight] = QSSGUtils::aux::translateLinearAttenuation(theLight->m_linearFade);
+                lightData.spotLightQuadAtt[spotLight] = QSSGUtils::aux::translateQuadraticAttenuation(theLight->m_quadraticFade);
+                float coneAngle = theLight->m_coneAngle;
+                // Inner cone angle must always be < cone angle, to not have possible undefined behavior for shader smoothstep
+                float innerConeAngle = std::min(theLight->m_innerConeAngle, coneAngle - 0.01f);
+                lightData.spotLightConeAngle[spotLight] = qDegreesToRadians(coneAngle);
+                lightData.spotLightInnerConeAngle[spotLight] = qDegreesToRadians(innerConeAngle);
+                spotLight++;
             }
+            theLightAmbientTotal += theLight->m_ambientColor;
         }
         // Copy light data
-        int lightOffset = shaders->offsetOfUniform("qt_pointLightPosition");
+        int lightOffset = shaders.offsetOfUniform("qt_pointLightPosition");
         if (lightOffset >= 0)
             memcpy(ubufData + lightOffset, &lightData, sizeof(ParticleLightData));
     }
-    shaders->setUniform(ubufData, "qt_light_ambient_total", &theLightAmbientTotal, 3 * sizeof(float), &cui.light_ambient_totalIdx);
+    shaders.setUniform(ubufData, "qt_light_ambient_total", &theLightAmbientTotal, 3 * sizeof(float), &cui.light_ambient_totalIdx);
     int enablePointLights = pointLight > 0 ? 1 : 0;
     int enableSpotLights = spotLight > 0 ? 1 : 0;
-    shaders->setUniform(ubufData, "qt_pointLights", &enablePointLights, sizeof(int));
-    shaders->setUniform(ubufData, "qt_spotLights", &enableSpotLights, sizeof(int));
+    shaders.setUniform(ubufData, "qt_pointLights", &enablePointLights, sizeof(int));
+    shaders.setUniform(ubufData, "qt_spotLights", &enableSpotLights, sizeof(int));
 
     // Line particle uniform
     int segmentCount = particleBuffer.segments();
     if (segmentCount) {
-        shaders->setUniform(ubufData, "qt_lineSegmentCount", &segmentCount, sizeof(int));
+        shaders.setUniform(ubufData, "qt_lineSegmentCount", &segmentCount, sizeof(int));
         float alphaFade = renderable.particles.m_alphaFade;
         float sizeModifier = renderable.particles.m_sizeModifier;
         float texcoordScale = renderable.particles.m_texcoordScale;
@@ -139,13 +146,13 @@ void QSSGParticleRenderer::updateUniformsForParticles(QSSGRef<QSSGRhiShaderPipel
             const auto size = image->m_texture.m_texture->pixelSize();
             texcoordScale *= float(size.height()) / float(size.width());
         }
-        shaders->setUniform(ubufData, "qt_alphaFade", &alphaFade, sizeof(float));
-        shaders->setUniform(ubufData, "qt_sizeModifier", &sizeModifier, sizeof(float));
-        shaders->setUniform(ubufData, "qt_texcoordScale", &texcoordScale, sizeof(float));
+        shaders.setUniform(ubufData, "qt_alphaFade", &alphaFade, sizeof(float));
+        shaders.setUniform(ubufData, "qt_sizeModifier", &sizeModifier, sizeof(float));
+        shaders.setUniform(ubufData, "qt_texcoordScale", &texcoordScale, sizeof(float));
     }
 }
 
-void QSSGParticleRenderer::updateUniformsForParticleModel(QSSGRef<QSSGRhiShaderPipeline> &shaderPipeline,
+void QSSGParticleRenderer::updateUniformsForParticleModel(QSSGRhiShaderPipeline &shaderPipeline,
                                                           char *ubufData,
                                                           const QSSGRenderModel *model,
                                                           quint32 offset)
@@ -153,11 +160,11 @@ void QSSGParticleRenderer::updateUniformsForParticleModel(QSSGRef<QSSGRhiShaderP
     auto &particleBuffer = *model->particleBuffer;
     const quint32 particlesPerSlice = particleBuffer.particlesPerSlice();
     const QVector2D oneOverSize = QVector2D(1.0f / particleBuffer.size().width(), 1.0f / particleBuffer.size().height());
-    shaderPipeline->setUniform(ubufData, "qt_oneOverParticleImageSize", &oneOverSize, 2 * sizeof(float));
-    shaderPipeline->setUniform(ubufData, "qt_countPerSlice", &particlesPerSlice, sizeof(quint32));
+    shaderPipeline.setUniform(ubufData, "qt_oneOverParticleImageSize", &oneOverSize, 2 * sizeof(float));
+    shaderPipeline.setUniform(ubufData, "qt_countPerSlice", &particlesPerSlice, sizeof(quint32));
     const QMatrix4x4 &particleMatrix = model->particleMatrix;
-    shaderPipeline->setUniform(ubufData, "qt_particleMatrix", &particleMatrix, 16 * sizeof(float));
-    shaderPipeline->setUniform(ubufData, "qt_particleIndexOffset", &offset, sizeof(quint32));
+    shaderPipeline.setUniform(ubufData, "qt_particleMatrix", &particleMatrix, 16 * sizeof(float));
+    shaderPipeline.setUniform(ubufData, "qt_particleIndexOffset", &offset, sizeof(quint32));
 }
 
 static void fillTargetBlend(QRhiGraphicsPipeline::TargetBlend &targetBlend, QSSGRenderParticles::BlendMode mode)
@@ -325,35 +332,36 @@ static QByteArray convertParticleData(QByteArray &dest, const QByteArray &data,
     return dest;
 }
 
-void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &shaderPipeline,
+void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPipeline,
+                                                QSSGPassKey passKey,
                                                 QSSGRhiContext *rhiCtx,
                                                 QSSGRhiGraphicsPipelineState *ps,
                                                 QSSGParticlesRenderable &renderable,
-                                                QSSGLayerRenderData &inData,
+                                                const QSSGLayerRenderData &inData,
                                                 QRhiRenderPassDescriptor *renderPassDescriptor,
                                                 int samples,
-                                                QSSGRenderCamera *camera,
-                                                int cubeFace,
+                                                int viewCount,
+                                                QSSGRenderCamera *alteredCamera,
+                                                QSSGRenderTextureCubeFace cubeFace,
                                                 QSSGReflectionMapEntry *entry)
 {
-    const void *layerNode = &inData.layer;
     const void *node = &renderable.particles;
     const bool needsConversion = !rhiCtx->rhi()->isTextureFormatSupported(QRhiTexture::RGBA32F);
 
-    QSSGRhiDrawCallData &dcd(cubeFace < 0 ? rhiCtx->drawCallData({ layerNode, node,
-                                                    nullptr, 0, QSSGRhiDrawCallDataKey::Main })
-                                          : rhiCtx->drawCallData({ layerNode, node,
-                                                                   entry, cubeFace, QSSGRhiDrawCallDataKey::Reflection }));
-    shaderPipeline->ensureUniformBuffer(&dcd.ubuf);
+    const auto cubeFaceIdx = QSSGBaseTypeHelpers::indexOfCubeFace(cubeFace);
+    QSSGRhiDrawCallData &dcd = QSSGRhiContextPrivate::get(rhiCtx)->drawCallData({ passKey, node, entry, cubeFaceIdx });
+    shaderPipeline.ensureUniformBuffer(&dcd.ubuf);
 
     char *ubufData = dcd.ubuf->beginFullDynamicBufferUpdateForCurrentFrame();
-    if (!camera)
-        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, *inData.camera);
-    else
-        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, *camera);
+    if (!alteredCamera) {
+        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, inData.renderedCameras);
+    } else {
+        QSSGRenderCameraList cameras({ alteredCamera });
+        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, cameras);
+    }
     dcd.ubuf->endFullDynamicBufferUpdateForCurrentFrame();
 
-    QSSGRhiParticleData &particleData = rhiCtx->particleData(&renderable.particles);
+    QSSGRhiParticleData &particleData = QSSGRhiContextPrivate::get(rhiCtx)->particleData(&renderable.particles);
     const QSSGParticleBuffer &particleBuffer = renderable.particles.m_particleBuffer;
     int particleCount = particleBuffer.particleCount();
     if (particleData.texture == nullptr || particleData.particleCount != particleCount) {
@@ -379,10 +387,10 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
 
     if (renderable.particles.m_depthSorting) {
         bool animatedParticles = renderable.particles.m_featureLevel == QSSGRenderParticles::FeatureLevel::Animated;
-        if (!camera)
-            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, *inData.cameraDirection, animatedParticles);
+        if (!alteredCamera)
+            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, inData.renderedCameraData.value()[0].direction, animatedParticles);
         else
-            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, camera->getScalingCorrectDirection(), animatedParticles);
+            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, alteredCamera->getScalingCorrectDirection(), animatedParticles);
         uploadData = convertParticleData(particleData.convertData, particleData.sortedData, needsConversion);
     } else {
         uploadData = convertParticleData(particleData.convertData, particleBuffer.data(), needsConversion);
@@ -395,11 +403,13 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
     rub->uploadTexture(particleData.texture, uploadDesc);
     rhiCtx->commandBuffer()->resourceUpdate(rub);
 
-    ps->ia.topology = QRhiGraphicsPipeline::TriangleStrip;
-    ps->ia.inputLayout = QRhiVertexInputLayout();
-    ps->ia.inputs.clear();
+    auto &ia = QSSGRhiInputAssemblerStatePrivate::get(*ps);
+    ia.topology = QRhiGraphicsPipeline::TriangleStrip;
+    ia.inputLayout = QRhiVertexInputLayout();
+    ia.inputs.clear();
 
     ps->samples = samples;
+    ps->viewCount = viewCount;
     ps->cullMode = QRhiGraphicsPipeline::None;
     if (renderable.renderableFlags.hasTransparency())
         fillTargetBlend(ps->targetBlend, renderable.particles.m_blendMode);
@@ -407,23 +417,23 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
         ps->targetBlend = QRhiGraphicsPipeline::TargetBlend();
 
     QSSGRhiShaderResourceBindingList bindings;
-    bindings.addUniformBuffer(0, VISIBILITY_ALL, dcd.ubuf, 0, shaderPipeline->ub0Size());
+    bindings.addUniformBuffer(0, VISIBILITY_ALL, dcd.ubuf, 0, shaderPipeline.ub0Size());
 
     // Texture maps
     // we only have one image
     QSSGRenderableImage *renderableImage = renderable.firstImage;
 
-    int samplerBinding = shaderPipeline->bindingForTexture("qt_sprite");
+    int samplerBinding = shaderPipeline.bindingForTexture("qt_sprite");
     if (samplerBinding >= 0) {
         QRhiTexture *texture = renderableImage ? renderableImage->m_texture.m_texture : nullptr;
         if (samplerBinding >= 0 && texture) {
             const bool mipmapped = texture->flags().testFlag(QRhiTexture::MipMapped);
-            QRhiSampler *sampler = rhiCtx->sampler({ toRhi(renderableImage->m_imageNode.m_minFilterType),
-                                                     toRhi(renderableImage->m_imageNode.m_magFilterType),
-                                                     mipmapped ? toRhi(renderableImage->m_imageNode.m_mipFilterType) : QRhiSampler::None,
-                                                     toRhi(renderableImage->m_imageNode.m_horizontalTilingMode),
-                                                     toRhi(renderableImage->m_imageNode.m_verticalTilingMode),
-                                                     QRhiSampler::Repeat
+            QRhiSampler *sampler = rhiCtx->sampler({ QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_minFilterType),
+                                                     QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_magFilterType),
+                                                     mipmapped ? QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_mipFilterType) : QRhiSampler::None,
+                                                     QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_horizontalTilingMode),
+                                                     QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_verticalTilingMode),
+                                                     QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_depthTilingMode)
                                                    });
             bindings.addTexture(samplerBinding, QRhiShaderResourceBinding::FragmentStage, texture, sampler);
         } else {
@@ -441,7 +451,7 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
         }
     }
 
-    samplerBinding = shaderPipeline->bindingForTexture("qt_particleTexture");
+    samplerBinding = shaderPipeline.bindingForTexture("qt_particleTexture");
     if (samplerBinding >= 0) {
         QRhiTexture *texture = particleData.texture;
         if (samplerBinding >= 0 && texture) {
@@ -456,7 +466,7 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
         }
     }
 
-    samplerBinding = shaderPipeline->bindingForTexture("qt_colorTable");
+    samplerBinding = shaderPipeline.bindingForTexture("qt_colorTable");
     if (samplerBinding >= 0) {
         bool hasTexture = false;
         if (renderable.colorTable) {
@@ -489,41 +499,44 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
         }
     }
 
+    QSSGRhiContextPrivate *rhiCtxD = QSSGRhiContextPrivate::get(rhiCtx);
+
     // TODO: This is identical to other renderables. Make into a function?
     QRhiShaderResourceBindings *&srb = dcd.srb;
     bool srbChanged = false;
     if (!srb || bindings != dcd.bindings) {
-        srb = rhiCtx->srb(bindings);
+        srb = rhiCtxD->srb(bindings);
+        rhiCtxD->releaseCachedSrb(dcd.bindings);
         dcd.bindings = bindings;
         srbChanged = true;
     }
 
-    if (cubeFace < 0)
+    if (cubeFace == QSSGRenderTextureCubeFaceNone)
         renderable.rhiRenderData.mainPass.srb = srb;
     else
-        renderable.rhiRenderData.reflectionPass.srb[cubeFace] = srb;
+        renderable.rhiRenderData.reflectionPass.srb[cubeFaceIdx] = srb;
 
-    const QSSGGraphicsPipelineStateKey pipelineKey = QSSGGraphicsPipelineStateKey::create(*ps, renderPassDescriptor, srb);
+    const auto pipelineKey = QSSGGraphicsPipelineStateKey::create(*ps, renderPassDescriptor, srb);
     if (dcd.pipeline
             && !srbChanged
             && dcd.renderTargetDescriptionHash == pipelineKey.extra.renderTargetDescriptionHash
             && dcd.renderTargetDescription == pipelineKey.renderTargetDescription
             && dcd.ps == *ps)
     {
-        if (cubeFace < 0)
+        if (cubeFace == QSSGRenderTextureCubeFaceNone)
             renderable.rhiRenderData.mainPass.pipeline = dcd.pipeline;
         else
             renderable.rhiRenderData.reflectionPass.pipeline = dcd.pipeline;
     } else {
-        if (cubeFace < 0) {
-            renderable.rhiRenderData.mainPass.pipeline = rhiCtx->pipeline(pipelineKey,
-                                                                          renderPassDescriptor,
-                                                                          srb);
+        if (cubeFace == QSSGRenderTextureCubeFaceNone) {
+            renderable.rhiRenderData.mainPass.pipeline = rhiCtxD->pipeline(pipelineKey,
+                                                                           renderPassDescriptor,
+                                                                           srb);
             dcd.pipeline = renderable.rhiRenderData.mainPass.pipeline;
         } else {
-            renderable.rhiRenderData.reflectionPass.pipeline = rhiCtx->pipeline(pipelineKey,
-                                                                          renderPassDescriptor,
-                                                                          srb);
+            renderable.rhiRenderData.reflectionPass.pipeline = rhiCtxD->pipeline(pipelineKey,
+                                                                                 renderPassDescriptor,
+                                                                                 srb);
             dcd.pipeline = renderable.rhiRenderData.reflectionPass.pipeline;
         }
         dcd.renderTargetDescriptionHash = pipelineKey.extra.renderTargetDescriptionHash;
@@ -532,12 +545,13 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRef<QSSGRhiShaderPipeline> &
     }
 }
 
-void QSSGParticleRenderer::prepareParticlesForModel(QSSGRef<QSSGRhiShaderPipeline> &shaderPipeline,
+void QSSGParticleRenderer::prepareParticlesForModel(QSSGRhiShaderPipeline &shaderPipeline,
                                                     QSSGRhiContext *rhiCtx,
                                                     QSSGRhiShaderResourceBindingList &bindings,
                                                     const QSSGRenderModel *model)
 {
-    QSSGRhiParticleData &particleData = rhiCtx->particleData(model);
+    QSSGRhiContextPrivate *rhiCtxD = QSSGRhiContextPrivate::get(rhiCtx);
+    QSSGRhiParticleData &particleData = rhiCtxD->particleData(model);
     const QSSGParticleBuffer &particleBuffer = *model->particleBuffer;
     int particleCount = particleBuffer.particleCount();
     bool update = particleBuffer.serial() != particleData.serial;
@@ -564,7 +578,7 @@ void QSSGParticleRenderer::prepareParticlesForModel(QSSGRef<QSSGRhiShaderPipelin
         rhiCtx->commandBuffer()->resourceUpdate(rub);
     }
     particleData.serial = particleBuffer.serial();
-    int samplerBinding = shaderPipeline->bindingForTexture("qt_particleTexture");
+    int samplerBinding = shaderPipeline.bindingForTexture("qt_particleTexture");
     if (samplerBinding >= 0) {
         QRhiTexture *texture = particleData.texture;
         if (samplerBinding >= 0 && texture) {
@@ -582,22 +596,24 @@ void QSSGParticleRenderer::prepareParticlesForModel(QSSGRef<QSSGRhiShaderPipelin
 
 void QSSGParticleRenderer::rhiRenderRenderable(QSSGRhiContext *rhiCtx,
                                                QSSGParticlesRenderable &renderable,
-                                               QSSGLayerRenderData &inData,
                                                bool *needsSetViewport,
-                                               int cubeFace,
-                                               QSSGRhiGraphicsPipelineState *state)
+                                               QSSGRenderTextureCubeFace cubeFace,
+                                               const QSSGRhiGraphicsPipelineState &state)
 {
     QRhiGraphicsPipeline *ps = renderable.rhiRenderData.mainPass.pipeline;
     QRhiShaderResourceBindings *srb = renderable.rhiRenderData.mainPass.srb;
 
-    if (cubeFace >= 0) {
+    if (cubeFace != QSSGRenderTextureCubeFaceNone) {
+        const auto cubeFaceIdx = QSSGBaseTypeHelpers::indexOfCubeFace(cubeFace);
         ps = renderable.rhiRenderData.reflectionPass.pipeline;
-        srb = renderable.rhiRenderData.reflectionPass.srb[cubeFace];
+        srb = renderable.rhiRenderData.reflectionPass.srb[cubeFaceIdx];
     }
 
     if (!ps || !srb)
         return;
 
+    Q_QUICK3D_PROFILE_START(QQuick3DProfiler::Quick3DRenderCall);
+
     QRhiCommandBuffer *cb = rhiCtx->commandBuffer();
     // QRhi optimizes out unnecessary binding of the same pipline
     cb->setGraphicsPipeline(ps);
@@ -605,10 +621,7 @@ void QSSGParticleRenderer::rhiRenderRenderable(QSSGRhiContext *rhiCtx,
     cb->setShaderResources(srb);
 
     if (needsSetViewport && *needsSetViewport) {
-        if (!state)
-            cb->setViewport(rhiCtx->graphicsPipelineState(&inData)->viewport);
-        else
-            cb->setViewport(state->viewport);
+        cb->setViewport(state.viewport);
         *needsSetViewport = false;
     }
     if (renderable.particles.m_featureLevel >= QSSGRenderParticles::FeatureLevel::Line) {
@@ -617,10 +630,12 @@ void QSSGParticleRenderer::rhiRenderRenderable(QSSGRhiContext *rhiCtx,
         int N = renderable.particles.m_particleBuffer.particleCount() / S;
         cb->draw(2 * S, N);
         QSSGRHICTX_STAT(rhiCtx, draw(2 * S, N));
+        Q_QUICK3D_PROFILE_END_WITH_ID(QQuick3DProfiler::Quick3DRenderCall, (2 * S | quint64(N) << 32), renderable.particles.profilingId);
     } else {
         // draw triangle strip with 2 triangles N times
         cb->draw(4, renderable.particles.m_particleBuffer.particleCount());
         QSSGRHICTX_STAT(rhiCtx, draw(4, renderable.particles.m_particleBuffer.particleCount()));
+        Q_QUICK3D_PROFILE_END_WITH_ID(QQuick3DProfiler::Quick3DRenderCall, (4 | quint64(renderable.particles.m_particleBuffer.particleCount()) << 32), renderable.particles.profilingId);
     }
 }
 
